// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.messageBus;

import com.microsoft.azure.servicebus.SubscriptionClient;
import com.microsoft.azure.servicebus.management.ManagementClient;
import com.microsoft.azure.servicebus.primitives.ServiceBusException;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.azure.cosmosdb.CosmosStore;
import org.opengroup.osdu.azure.partition.PartitionServiceClient;
import org.opengroup.osdu.azure.serviceBusManager.IManagementClientFactory;
import org.opengroup.osdu.core.common.model.notification.HmacSecret;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.notification.ISubscriptionFactory;
import org.opengroup.osdu.core.common.provider.interfaces.ITenantFactory;
import org.opengroup.osdu.notification.provider.azure.messageBus.ProcessNotification;
import org.opengroup.osdu.notification.provider.azure.messageBus.SubscriptionClientFactImpl;
import org.opengroup.osdu.notification.provider.azure.messageBus.SubscriptionManagerImpl;
import org.opengroup.osdu.notification.provider.azure.messageBus.models.TopicSubscriptions;
import org.opengroup.osdu.notification.provider.azure.messageBus.thread.ThreadDpsHeaders;
import org.opengroup.osdu.notification.provider.azure.util.AzureCosmosProperties;
import org.opengroup.osdu.notification.provider.azure.util.AzureServiceBusConfig;

import java.util.Collections;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import static java.lang.Thread.sleep;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.atLeast;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.lenient;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@RunWith(MockitoJUnitRunner.class)
public class SubscriptionManagerImplTest {

    private static final String maxLockRenewDuration = "60";
    private static final String maxConcurrentCalls = "1";
    private static final String nThreads = "2";
    private static final String errorMessage = "some-error";

    @InjectMocks
    private SubscriptionManagerImpl subscriptionManager;
    @Mock
    private SubscriptionClientFactImpl subscriptionClientFactory;
    @Mock
    private TopicSubscriptions topicSubscriptions;
    @Mock
    private SubscriptionClient subscriptionClient;
    @Mock
    private ProcessNotification processNotification;
    @Mock
    private ITenantFactory tenantFactory;
    @Mock
    private ISubscriptionFactory subscriptionFactory;
    @Mock
    private ThreadDpsHeaders dpsHeaders;
    @Mock
    private PartitionServiceClient partitionService;
    @Mock
    private IManagementClientFactory factory;
    @Mock
    private ManagementClient managementClient;
    @Mock
    private CosmosStore cosmosStore;
    @Mock
    private AzureCosmosProperties azureCosmosProperties;
    @Mock
    private AzureServiceBusConfig azureServiceBusConfig;

    private static final String dataPartition = "testTenant";
    private static final String cosmosDbName = "testDatabase";
    private static final String registerContainerName = "testContainer";
    private static final String sbTopic = "testTopic";
    private static final String sbSubscription = "testSubscription";
    private static final String initial_thread_delay = "0";
    private static final String thread_delay = "1";

    @Before
    public void init() throws ServiceBusException, InterruptedException {
        MockitoAnnotations.openMocks(SubscriptionManagerImplTest.this);
        TenantInfo tenantInfo = new TenantInfo();
        tenantInfo.setDataPartitionId(dataPartition);
        lenient().when(azureServiceBusConfig.getMaxConcurrentCalls()).thenReturn(maxConcurrentCalls);
        lenient().when(azureServiceBusConfig.getNThreads()).thenReturn(nThreads);
        lenient().when(azureServiceBusConfig.getMaxLockRenewDurationInSeconds()).thenReturn(maxLockRenewDuration);
        lenient().when(azureServiceBusConfig.getInitialSubscriptionManagerDelay()).thenReturn(initial_thread_delay);
        lenient().when(azureServiceBusConfig.getConsecutiveSubscriptionManagerDelay()).thenReturn(thread_delay);
        lenient().when(azureCosmosProperties.registerSubscriptionContainerName()).thenReturn(registerContainerName);
        lenient().when(azureCosmosProperties.cosmosDBName()).thenReturn(cosmosDbName);
        lenient().when(tenantFactory.listTenantInfo()).thenReturn(Collections.singletonList(tenantInfo));
        lenient().when(factory.getManager(anyString())).thenReturn(managementClient);
        lenient().when(managementClient.topicExists(anyString())).thenReturn(true, false);
        lenient().when(managementClient.subscriptionExists(anyString(), anyString())).thenReturn(true, false);
        lenient().when(cosmosStore.findAllItems(dataPartition, cosmosDbName, registerContainerName, Subscription.class)).
                thenReturn(Collections.singletonList(getHmac_subscription()));
    }

    @Test
    public void shouldSuccessfullyRegisterMessageHandler() throws ServiceBusException, InterruptedException {

        subscriptionManager.subscribeNotificationsEvent();

        verify(azureServiceBusConfig, times(1)).getNThreads();
        verify(azureCosmosProperties, times(1)).registerSubscriptionContainerName();
        verify(azureCosmosProperties, times(1)).cosmosDBName();
    }

    @Test
    public void shouldNotRegisterMessageHandlerWhenAlreadyRegisteredInPrevThread() throws ServiceBusException, InterruptedException {

        when(topicSubscriptions.checkIfNewTopicSubscription(anyString(), anyString(), anyString())).thenReturn(true);
        when(subscriptionClientFactory.getSubscriptionClient(dataPartition, sbTopic, sbSubscription)).thenReturn(subscriptionClient);
        doNothing().when(subscriptionClient).registerMessageHandler(any(), any(), any());
        Executors.newSingleThreadScheduledExecutor().scheduleAtFixedRate(subscriptionManager, Integer.parseUnsignedInt(azureServiceBusConfig.getInitialSubscriptionManagerDelay()),
                Integer.parseUnsignedInt(azureServiceBusConfig.getConsecutiveSubscriptionManagerDelay()), TimeUnit.SECONDS);
        sleep(10000);
        verify(azureServiceBusConfig, times(1)).getMaxConcurrentCalls();
        verify(azureServiceBusConfig, atLeast(2)).getNThreads();
        verify(azureServiceBusConfig, times(1)).getMaxLockRenewDurationInSeconds();
        verify(azureCosmosProperties, atLeast(2)).registerSubscriptionContainerName();
        verify(azureCosmosProperties, atLeast(2)).cosmosDBName();
        verify(subscriptionClientFactory, times(1)).getSubscriptionClient(dataPartition, sbTopic, sbSubscription);
        verify(subscriptionClient, times(1)).registerMessageHandler(any(), any(), any());

    }

    @Test
    public void shouldNotRegisterMessageHandlerIfSSubscriptionClientThrowsException() throws ServiceBusException, InterruptedException {
        lenient().when(topicSubscriptions.checkIfNewTopicSubscription(anyString(), anyString(), anyString())).thenReturn(true);
        lenient().doNothing().when(subscriptionClient).registerMessageHandler(any(), any(), any());
        doThrow(new InterruptedException(errorMessage)).when(subscriptionClientFactory).getSubscriptionClient(dataPartition, sbTopic, sbSubscription);
        subscriptionManager.subscribeNotificationsEvent();

        verify(azureServiceBusConfig, times(0)).getMaxConcurrentCalls();
        verify(azureServiceBusConfig, times(1)).getNThreads();
        verify(azureServiceBusConfig, times(0)).getMaxLockRenewDurationInSeconds();
        verify(azureCosmosProperties, times(1)).registerSubscriptionContainerName();
        verify(azureCosmosProperties, times(1)).cosmosDBName();
        verify(subscriptionClientFactory, times(1)).getSubscriptionClient(dataPartition, sbTopic, sbSubscription);
        verify(subscriptionClient, times(0)).registerMessageHandler(any(), any(), any());
    }

    @Test
    public void shouldThrowExceptionIfErrorWhileRegisteringMessageHandler() throws ServiceBusException, InterruptedException {

        lenient().when(topicSubscriptions.checkIfNewTopicSubscription(anyString(), anyString(), anyString())).thenReturn(true);
        doThrow(new InterruptedException(errorMessage)).when(subscriptionClient).registerMessageHandler(any(), any(), any());
        when(subscriptionClientFactory.getSubscriptionClient(dataPartition, sbTopic, sbSubscription)).thenReturn(subscriptionClient);
        subscriptionManager.subscribeNotificationsEvent();

        verify(azureServiceBusConfig, times(1)).getMaxConcurrentCalls();
        verify(azureServiceBusConfig, times(1)).getNThreads();
        verify(azureServiceBusConfig, times(1)).getMaxLockRenewDurationInSeconds();
        verify(azureCosmosProperties, times(1)).registerSubscriptionContainerName();
        verify(azureCosmosProperties, times(1)).cosmosDBName();
        verify(subscriptionClientFactory, times(1)).getSubscriptionClient(dataPartition, sbTopic, sbSubscription);
        verify(subscriptionClient, times(1)).registerMessageHandler(any(), any(), any());
    }

    private static Subscription getHmac_subscription() {
        Subscription hmac_subscription = new Subscription();
        hmac_subscription.setName("hamc_test_subscription");
        hmac_subscription.setPushEndpoint("http://challenge");
        hmac_subscription.setDescription("Description");
        hmac_subscription.setTopic(sbTopic);
        hmac_subscription.setNotificationId(sbSubscription);
        hmac_subscription.setId("id_1");
        hmac_subscription.setCreatedBy("test@test.com");
        HmacSecret secret = new HmacSecret();
        secret.setValue("testsecret");
        hmac_subscription.setSecret(secret);
        return hmac_subscription;
    }
}
